/* Convert an OFF file to PLG format */

/* Written by Bernie Roehl, January 1992 */

#include <stdio.h>
#include <alloc.h>
#include <string.h>
#include <math.h>

double scale = 10.0;

void main(int argc, char *argv[])
	{
	int i;
	if (argc < 2) {
		fprintf(stderr, "Usage: off2plg [-sscale] infile [infile ...]\n");
		exit(1);
		}
	while (argc > 1) {
		if (*argv[1] == '-')
			switch (argv[1][1]) {
				case 's': scale = atof(&argv[1][2]); break;
				default: printf("Ignoring invalid option '%s'\n", argv[1]); break;
				}
		else
			process(argv[1]);
		--argc; ++argv;
		}
	}

void synerr(int line, char *filename)
	{
	printf("Syntax error in line %d of file '%s'\n", line, filename);
	}

#define MAXPTS 100     /* up to 100-sided polys can be handled */
int polyverts[MAXPTS];

int *colormap = NULL;  /* mapping of polygon number to color number */
int ncolors = 0;

process(char *fn)
	{
	char filename[80], *colorfile = NULL, *geofile = NULL, *name;
	char *author = NULL, *copyright = NULL, *description = NULL;
	FILE *in, *out;
	int nv, np, i;
	double minval = 0, maxval = 0;
	enum { CW, CCW } order = CW;
	enum { DEF, GENERIC, INDEXED, UNKNOWN } colortype = UNKNOWN;
	name = fn;
	printf("Processing OFF file '%s'...\n", fn);
	sprintf(filename, "%s.plg", fn);
	if ((out = fopen(filename, "w")) == NULL) {
		printf("Could not create PLG output file '%s'\n");
		return -1;
		}
	/* attempt to read the AOFF file, if it exists */
	sprintf(filename, "%s.aoff", fn);
	if ((in = fopen(filename, "r")) != NULL) {
		char buff[100];
		int line = 0;
		printf("Reading '%s'...\n", filename);
		while (fgets(buff, sizeof(buff), in)) {
			char propname[100], proptype[100], propformat[100], propdata[100];
			int nparm;
			++line;
			if (buff[0] == '\n' || buff[0] == '#') continue;
			if ((nparm = sscanf(buff, "%s %s %s %s", propname, proptype, propformat, propdata)) < 2) {
				synerr(line, filename);
				continue;
				}
			if (!stricmp(propname, "name")) name = strdup(proptype);
			else if (!stricmp(propname, "type")) {
				if (stricmp(proptype, "polygon")) {
					printf("skipping '%s': not type 'polygon'\n", fn);
					fclose(in);
					fclose(out);
					return -1;
					}
				}
			else if (!stricmp(propname, "author"))
				author = strdup(buff);
			else if (!stricmp(propname, "copyright"))
				copyright = strdup(buff);
			else if (!stricmp(propname, "description"))
				description = strdup(buff);
			else if (!stricmp(propname, "vertex_order")) {
				if (nparm != 4) {
					synerr(line, filename);
					continue;
					}
				if (stricmp(propformat, "s")) {
					printf("Unknown format for vertex_order in line %d of file '%s'\n", line, filename);
					continue;
					}
				if (!stricmp(propdata, "clockwise")) order = CW;
				else if (!stricmp(propdata, "counter-clockwise")) order = CCW;
				else if (!stricmp(propdata, "counterclockwise")) order = CCW;
				else {
					printf("Unknown vertex ordering in line %d of file '%s'\n", line, filename);
					continue;
					}
				}
			else if (!stricmp(propname, "polygon_colors")) {
				if (nparm != 4) {
					synerr(line, filename);
					continue;
					}
				if (stricmp(propformat, "fff")) {
					printf("Color format not 'fff' in line %d of file '%s'\n", line, filename);
					continue;
					}
				if (!stricmp(proptype, "default")) colortype = DEF;
				else if (!stricmp(proptype, "generic")) colortype = GENERIC;
				else if (!stricmp(proptype, "indexed")) colortype = INDEXED;
				else {
					printf("Unknown color type '%s' in line %s of file '%s'\n", proptype, line, filename);
					continue;
					}
				colorfile = strdup(propdata);
				}
			else if (!stricmp(propname, "geometry")) {
				if (nparm != 4) {
					synerr(line, filename);
					continue;
					}
				if (stricmp(propformat, "fff")) {
					printf("Geometry format not 'fff' in line %f of file '%s'\n", line, filename);
					printf("Skipping file '%s'\n", fn);
					fclose(out);
					return -2;
					}
				if (stricmp(proptype, "indexed_poly")) {
					printf("Geometry type not 'indexed_poly' in line %f of file '%s'\n", line, filename);
					printf("Skipping file '%s'\n", fn);
					fclose(out);
					return -3;
					}
				geofile = strdup(propdata);
				}
			}
		fclose(in);
		}
	else
		printf("File '%s' not found; using defaults...\n", filename);
	printf("Vertex order = %s\n", (order == CW) ? "clockwise" : "counter-clockwise");
	/* attempt to process the color file, if it exists */
	switch (colortype) {
		case GENERIC: load_generic(colorfile); break;
		case INDEXED: load_indexed(colorfile); break;
		case UNKNOWN:
			sprintf(filename, "%s.ipcol", fn);
			if ((in = fopen(filename, "r")) != NULL) {
				fclose(in);
				load_indexed(filename);
				break;
				}
			sprintf(filename, "%s.pcol", fn);
			if ((in = fopen(filename, "r")) != NULL) {
				fclose(in);
				load_generic(filename);
				break;
				}
			/* else fall through */
		case DEF: break;  /* no colors specified */
		}
	/* attempt to process the geometry file */
	if (geofile == NULL) {
		sprintf(filename, "%s.geom", fn);
		geofile = filename;
		}
	if ((in = fopen(geofile, "r")) == NULL) {
		printf("Could not open geometry file '%s'\n", geofile);
		fclose(out);
		return -4;
		}
	if (fscanf(in, "%d %d %*d", &nv, &np) != 2) {
		synerr(1, geofile);
		fclose(in);
		fclose(out);
		return -5;
		}
	printf("Scanning vertices...\n");
	for (i = 0; i < nv; ++i) {  /* scan through the vertices */
		double x, y, z;
		fscanf(in, "%lf %lf %lf", &x, &y, &z);
		if (i == 0) minval = maxval = x;
		if (x < minval) minval = x; if (y < minval) minval = y; if (z < minval) minval = z;
		if (x < maxval) maxval = x; if (y < maxval) maxval = y; if (z < maxval) maxval = z;
		}
	rewind(in);  /* go back to re-read vertices */
	fscanf(in, "%*d %*d %*d");  /* skip the header */
	if (description) fprintf(out, "# %s\n", description);
	if (author) fprintf(out, "# %s\n", author);
	if (copyright) fprintf(out, "# %s\n", copyright);
	fprintf(out, "%s %d %d\n", name, nv, np);  /* write out PLG header line */
	printf("Scaling and generating vertices...\n");
	while (nv--) {  /* read and output the vertices, scaling and integerizing them */
		double x, y, z;
		fscanf(in, "%lf %lf %lf", &x, &y, &z);
#ifdef OLD
		fprintf(out, "%f %f %f\n", x, y, z);
#endif
		fprintf(out, "%ld %ld %ld\n", (long) floor((x*scale)/maxval), (long) floor((y*scale)/maxval), (long) floor((z*scale)/maxval));
		}
	printf("Processing polygon data...\n");
	for (i = 0; i < np; ++i) {  /* process the polygons */
		int npt, j;
		fscanf(in, "%d", &npt);
		if (npt >= MAXPTS) {
			printf("Too many vertices in poly %d in file '%s'\n", i + 1, geofile);
			fclose(in);
			fclose(out);
			return -8;
			}
		for (j = 0; j < npt; ++j)
			fscanf(in, "%d", &polyverts[j]);
		if (ncolors) fprintf(out, "%d %d", colormap[(i >= ncolors) ? ncolors : i], npt);
		else fprintf(out, "1 %d", npt);
		if (order == CW)
			for (j = 0; j < npt; ++j)
				fprintf(out, " %d", polyverts[j]-1);
		else  /* CCW */
			for (j = npt-1; j >= 0; --j)
				fprintf(out, " %d", polyverts[j]-1);
		fprintf(out, "\n");
		}
	printf("All done '%s'\n", fn);
	fclose(in);
	fclose(out);
	if (colormap) free(colormap);  /* tidy up */
	colormap = NULL; ncolors = 0;
	return 0;
	}

load_indexed(char *filename)
	{
	FILE *in;
	int nmap, i;
	printf("Processing indexed color file '%s'\n", filename);
	if ((in = fopen(filename, "r")) == NULL) {
		printf("Could not open '%s'\n", filename);
		return -1;
		}
	fscanf(in, "%d %d", &nmap, &ncolors);
	if ((colormap = calloc(ncolors, sizeof(int))) == NULL) {
		printf("Could not allocate %d colors in colormap\n", ncolors);
		fclose(in);
		return -2;
		}
	while (nmap--) fscanf(in, "%*f %*f %*f");  /* skip actual colors */
	for (i = 0; i < ncolors; ++i) {
		fscanf(in, "%d", &colormap[i]);
		--colormap[i];  /* change to origin zero */
		}
	fclose(in);
	return 0;
	}

typedef struct { float r, g, b; } COLOR;

load_generic(char *filename)
	{
	FILE *in;
	COLOR *map;
	int nmap, i;
	printf("Processing generic color file '%s'\n", filename);
	if ((in = fopen(filename, "r")) == NULL) {
		printf("Could not open '%s'\n", filename);
		return -1;
		}
	fscanf(in, "%d", &nmap);
	if ((colormap = calloc(nmap, sizeof(int))) == NULL) {
		printf("Could not allocate %d colors in colormap\n", nmap);
		fclose(in);
		return -2;
		}
	if ((map = calloc(nmap, sizeof(COLOR))) == NULL) {
		printf("Could not allocate %d colors in holding array\n", nmap);
		fclose(in);
		return -3;
		}
	for (i = ncolors = 0; i < nmap; ++i) {
		float r, g, b;
		int j;
		fscanf(in, "%f %f %f", &r, &g, &b);
		for (j = 0; j < i; ++j)
			if (map[j].r == r && map[j].g == g && map[j].b == b)
				break;
		if (j < i) colormap[i] = j;
		else {
			map[ncolors].r = r; map[ncolors].g = g; map[ncolors].b = b;
			colormap[i] = ncolors++;
			}
		}		
	free(map);
	fclose(in);
	return 0;
	}
